import datetime as dt
import math
import re
from typing import Optional

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from dateutil.relativedelta import relativedelta
from django.db.models import ExpressionWrapper, F, FloatField, QuerySet
from pandas.tseries.offsets import BYearEnd
from plotly.subplots import make_subplots
from wbcore.contrib.currency.models import CurrencyFXRates

from wbfdm.enums import MarketData
from wbfdm.models import Instrument, InstrumentPrice


class FinancialsChartGenerator:
    def __init__(self, instrument: Instrument):
        self.instrument = instrument

    def _make_presentable_fields(self, fields_list: list, bold: bool = True) -> pd.Index:
        full_capitalize_fields = self.get_full_capitalize_fields()
        percent_fields = self.get_percent_fields()
        finance_friendly = self._rename_fields_into_finance_friendly()
        pattern = re.compile(r"\b(" + "|".join(finance_friendly.keys()) + r")\b")

        for i, x in enumerate(fields_list):
            y = pattern.sub(lambda z: finance_friendly[z.group()], x)
            if x in percent_fields:
                x = y + "_(%)"
            else:
                x = y
            fields_list[i] = x

        fields_list = [x.split("_") for x in fields_list]
        for i, primary_list in enumerate(fields_list):
            for j, word in enumerate(primary_list):
                primary_list[j] = word.upper() if word in full_capitalize_fields else word.capitalize()

            fields_list[i] = " ".join(primary_list)
        fields_list = [f"<b>{x}</b>" for x in fields_list] if bold else fields_list
        return pd.Index(fields_list)

    def _make_pretty_table(self, df: pd.DataFrame) -> go.Figure:
        previous_years = [elem for elem in df.columns if elem.find("E") == -1]
        big_int_fields = self.get_big_int_fields()
        percent_fields = self.get_percent_fields()
        multiples_fields = self.get_multiples_fields()
        big_int_fields = df.index.intersection(big_int_fields)
        percent_fields = df.index.intersection(percent_fields)
        multiples_fields = df.index.intersection(multiples_fields)
        other_fields = df.index.difference(big_int_fields.append(percent_fields).append(multiples_fields))
        df.loc[big_int_fields] = (
            df.loc[big_int_fields].div(1000000).applymap(lambda x: f"{x:,.0f}" if not np.isnan(x) else "")
        )
        df.loc[percent_fields] = df.loc[percent_fields].applymap(lambda x: f"{x:,.1%}" if not np.isnan(x) else "")
        df.loc[multiples_fields] = df.loc[multiples_fields].applymap(lambda x: f"{x:,.1f}x" if not np.isnan(x) else "")
        df.loc[other_fields] = df.loc[other_fields].applymap(lambda x: f"{x:,.2f}" if not np.isnan(x) else "")
        df.insert(
            0,
            f"{self.instrument.name_repr} in USD mn",
            self._make_presentable_fields(fields_list=df.index.tolist()),
        )
        df_table = df.reset_index().drop("index", axis=1)
        df_table.columns = df_table.columns.map("<b>{}</b>".format)
        colors = ["white", "#EEEEF1"] * math.floor(len(df_table.index) / 2)
        colors += ["white"] if len(df_table.index) % 2 == 1 else ""
        fig = go.Figure(
            go.Table(
                columnwidth=[3] + [1] * len(df_table.columns),
                header=dict(
                    values=df_table.columns.tolist(),
                    line_color="darkslategray",
                    fill_color="grey",
                    font=dict(color="white", size=13),
                    align=["left", "center"],
                    height=40,
                ),
                cells=dict(
                    values=df.T.values.tolist(),
                    align=["left", "center"],
                    line_color="darkslategray",
                    fill_color=["lightgrey"] + [colors] * len(previous_years) + ["lightyellow"],
                    height=30,
                ),
            ),
        )
        return fig

    @staticmethod
    def _rename_fields_into_finance_friendly() -> dict:
        return {
            "gross_profit_margin": "gross_profit_margin_[non-gaap]",
            "ebitda_margin": "ebitda_margin_[non-gaap]",
            "net_profit_margin": "net_profit_margin_[gaap]",
            "net_profit": "net_profit_[non-gaap]",
            "reported_net_profit": "net_profit_[gaap]",
            "eps": "eps_[non-gaap]",
            "eps_growth": "eps_growth_[non-gaap]",
        }

    @staticmethod
    def get_big_int_fields() -> list[str]:
        return [
            "revenue",
            "cost_of_good_sold",
            "gross_profit",
            "ebitda",
            "ebit",
            "net_profit",
            "reported_net_profit",
            "cost_research_development",
            "sga",
            "free_cash_flow",
            "cash_from_operation",
            "working_capital",
            "capital_expenditures",
            "investment_cash",
            "financing_cash",
            "shareholder_equity",
            "total_assets",
            "current_liabilities",
            "total_debt",
            "cash_and_cash_equivalents",
            "net_debt",
            "cash_and_short_term_investments",
            "net_change_in_cash",
            "receivables",
            "inventories",
            "payables",
            "current_assets",
            "entreprise_value",
            "free_cash",
            "burn_rate",
            "operating_burn_rate",
            "free_cash_burn_rate",
            "operating_cash_flow",
            "investing_cash_flow",
            "unlevered_free_cash_flow",
        ]

    @staticmethod
    def get_percent_fields() -> list[str]:
        return [
            "revenue_growth",
            "revenue_growth_5y_cagr",
            "gross_profit_margin",
            "ebitda_margin",
            "ebit_margin",
            "net_profit_margin",
            "free_cash_flow_growth",
            "return_on_equity",
            "return_on_assets",
            "return_on_capital_employed",
            "operating_cash_flow_growth",
            "investing_cash_flow_growth",
            "unlevered_free_cash_flow_growth",
            "unlevered_free_cash_flow_margin",
            "diluted_eps_growth",
            "eps_growth",
        ]

    @staticmethod
    def get_full_capitalize_fields() -> list[str]:
        return ["ebitda", "ebit", "eps", "sga", "ytd", "[gaap]", "[non-gaap]"]

    @staticmethod
    def get_multiples_fields() -> list[str]:
        return ["net_debt_to_ebitda_ratio"]

    def get_latest_year(self, queryset: Optional[QuerySet] = None) -> int:
        # if queryset is None:
        #     fundamentals = Fundamental.annual_objects.filter(instrument=self.instrument)  # noqa
        #     if fundamentals.exists():
        #         return fundamentals.latest("period__date_range").period.period_year
        # elif queryset.exists():  # not empty
        #     return queryset.latest("period__date_range").period.period_year

        return dt.date.today().year - 1  # if no data provided, take last previous year

    def fundamentals_df(self, from_n_years_before: int = 3) -> pd.DataFrame:
        return pd.DataFrame()
        # fundamentals_fields_list = Fundamental.get_number_serializer_fields().keys()  # noqa
        # qs_fundamentals = Fundamental.annual_objects.filter(instrument=self.instrument)  # noqa
        # latest_year = self.get_latest_year(queryset=qs_fundamentals)
        # qs_fundamentals = (
        #     qs_fundamentals.filter(
        #         period__period_year__gte=latest_year - from_n_years_before,
        #         period__period_year__lte=latest_year,
        #     )
        #     .order_by("period__period_year")
        #     .values(*Fundamental.get_number_serializer_fields().keys(), "period__period_year")  # noqa
        # )
        # df_fundamentals = pd.DataFrame(qs_fundamentals)
        # if not df_fundamentals.empty:
        #     df_fundamentals["period__period_year"] = df_fundamentals.period__period_year.astype(str)
        #     df_fundamentals = df_fundamentals.set_index("period__period_year")[fundamentals_fields_list].T
        #     return df_fundamentals
        # return df_fundamentals.reindex(fundamentals_fields_list)

    def forecasts_df(self) -> pd.DataFrame:
        return pd.DataFrame()
        # forecasts_fields_list = list(Forecast.get_number_serializer_fields().keys())  # noqa
        # latest_year = self.get_latest_year(queryset=None)
        # forecasts = self.instrument.forecasts.filter(revenue_y1__isnull=False)
        # if forecasts.exists():
        #     last_forecast = forecasts.latest("date")
        #     df_forecasts = pd.DataFrame.from_dict(model_to_dict(last_forecast), orient="index").T[
        #         forecasts_fields_list
        #     ]
        # else:
        #     df_forecasts = pd.DataFrame(index=forecasts_fields_list)
        # forecasts_fields_list += ["reported_net_profit"]
        #
        # def split_year_fields(_year_n: str) -> list:
        #     return [field for field in forecasts_fields_list if field.find(_year_n) != -1]
        #
        # next_years = list(str(year) + "E" for year in range(latest_year + 1, latest_year + 6))
        # df_forecasts_ordered = pd.DataFrame(index=forecasts_fields_list)
        #
        # if forecasts.exists():
        #     for year, e_year in zip(["_y1", "_y2", "_y3", "_y4", "_y5"], next_years):
        #         year_fields = split_year_fields(_year_n=year)
        #         index = list(map(lambda x: x.replace(year, ""), year_fields))
        #         tmp = pd.DataFrame(index=[e_year], columns=index, data=df_forecasts[year_fields].values.tolist()).T
        #         df_forecasts_ordered = pd.concat([df_forecasts_ordered, tmp], axis=1)
        # return df_forecasts_ordered

    def combine_fundamentals_and_forecasts_df(self) -> pd.DataFrame:
        df_fundamentals = self.fundamentals_df()
        df_forecasts = self.forecasts_df()

        # Hardcode rename for those which do not have same field name:
        df_forecasts.rename(index={"adjusted_net_profit": "net_profit"}, inplace=True)

        df_table = pd.concat([df_fundamentals, df_forecasts], axis=1).astype(float)

        return df_table

    def get_instrument_price_data(
        self,
        from_date: dt.date | None = None,
        to_date: dt.date | None = None,
    ) -> pd.DataFrame:
        if not from_date:
            from_date = dt.date(dt.date.today().year - 1, 1, 1)

        df_prices = pd.DataFrame(
            Instrument.objects.filter(id=self.instrument.id).dl.market_data(
                values=[MarketData.CLOSE, MarketData.OPEN, MarketData.LOW, MarketData.HIGH, MarketData.VOLUME],
                from_date=from_date,
                to_date=to_date,
            )
        )
        if df_prices.empty:
            return pd.DataFrame()
        df_prices = df_prices[
            [
                MarketData.CLOSE.value,
                MarketData.OPEN.value,
                MarketData.LOW.value,
                MarketData.HIGH.value,
                MarketData.VOLUME.value,
                "valuation_date",
            ]
        ].rename(columns={"valuation_date": "date"})
        df_prices = df_prices.set_index("date").sort_index().astype(float)
        timeline = pd.date_range(df_prices.index[0], df_prices.index[-1])
        df_prices = df_prices.reindex(timeline, method="ffill")
        return df_prices[df_prices["close"] != 0]

    def summary_table_chart(self) -> go.Figure:
        df_table = self.combine_fundamentals_and_forecasts_df()

        if not df_table.empty:
            df_table.loc["reported_net_profit", :].replace([np.nan, np.inf, -np.inf], None, inplace=True)
            df_table = df_table.loc[
                [
                    "revenue",
                    "revenue_growth",
                    "gross_profit_margin",
                    "ebitda_margin",
                    "reported_net_profit",
                    "net_profit_margin",
                    "net_profit",
                    "eps",
                    "return_on_equity",
                    "return_on_assets",
                    "return_on_capital_employed",
                    "return_on_invested_capital",
                    "net_debt_to_ebitda_ratio",
                    "interest_coverage_ratio",
                    "employee_count",
                    "employee_count_growth",
                ],
                :,
            ]
            df_table.loc["eps_growth", :] = df_table.loc["eps", :].pct_change()
            df_table = df_table.reindex(
                [
                    "revenue",
                    "revenue_growth",
                    "gross_profit_margin",
                    "ebitda_margin",
                    "reported_net_profit",
                    "net_profit_margin",
                    "net_profit",
                    "eps",
                    "eps_growth",
                    "return_on_equity",
                    "return_on_assets",
                    "return_on_capital_employed",
                    "return_on_invested_capital",
                    "net_debt_to_ebitda_ratio",
                    "interest_coverage_ratio",
                    "employee_count",
                    "employee_count_growth",
                ]
            )

        fig = self._make_pretty_table(df=df_table)
        return fig

    @staticmethod
    def replace_hovertemplate(fig: go.Figure, i_position: int, text: str) -> go.Figure:
        try:
            fig["data"][i_position]["hovertemplate"] = text
        except IndexError:
            pass
        return fig

    def financials_chart(self) -> go.Figure:
        fig = make_subplots(specs=[[{"secondary_y": True}]])
        df_table = self.combine_fundamentals_and_forecasts_df()
        if not df_table.empty:
            df_table = df_table.loc[["revenue", "net_profit", "ebitda_margin", "net_profit_margin"], :]

            for variable in ["revenue", "net_profit"]:
                text = "Revenue" if variable == "revenue" else "Net Profit"
                fig.add_bar(
                    y=df_table.loc[variable, :].dropna().div(1000000).round(2).values,
                    x=df_table.loc[variable, :].dropna().index,
                    name="Revenue" if variable == "revenue" else "Net Profit",
                    yaxis="y",
                    hovertemplate="<b>" + text + "</b><br>%{x}<br>%{y:.2f} Mio.<extra></extra>",
                )
            for variable in ["ebitda_margin", "net_profit_margin"]:
                text = "EBITDA Margin" if variable == "ebitda_margin" else "Net Profit Margin"
                fig.add_scatter(
                    y=df_table.loc[variable, :].dropna().mul(100).values,
                    x=df_table.loc[variable, :].dropna().index,
                    name=text,
                    yaxis="y2",
                    hovertemplate="<b>" + text + "</b><br>%{x}<br>%{y:.2f}%<extra></extra>",
                )
        fig.update_xaxes(nticks=len(df_table.columns))
        symbol = self.instrument.currency.symbol
        fig.update_yaxes(title=f"{symbol if symbol else ''} in Million", secondary_y=False)
        fig.update_yaxes(ticksuffix="%", secondary_y=True)
        return fig

    def profitability_ratios_chart(self) -> go.Figure:
        df_table = self.combine_fundamentals_and_forecasts_df()
        if not df_table.empty:
            df_table = df_table.loc[
                ["return_on_equity", "return_on_assets", "return_on_capital_employed", "return_on_invested_capital"]
            ]

            df_table.rename(
                index={
                    "return_on_equity": "ROE",
                    "return_on_assets": "ROA",
                    "return_on_capital_employed": "ROCE",
                    "return_on_invested_capital": "ROIC",
                },
                inplace=True,
            )
        fig = go.Figure()
        for index_name in df_table.index:
            fig.add_trace(
                go.Scatter(
                    x=df_table.loc[index_name, :].dropna().index,
                    y=df_table.loc[index_name, :].mul(100).dropna().values,
                    mode="lines+markers",
                    name=index_name,
                    hovertemplate=f"<b>{index_name}</b>" + "<br>%{y:.2f}%<extra></extra>",
                )
            )
        fig.update_yaxes(ticksuffix="%")
        fig.update_layout(hovermode="x")
        return fig

    def stock_performance_summary_chart(self) -> go.Figure:
        df_prices = self.get_instrument_price_data()
        if df_prices.empty:
            return go.Figure()
        last_day = df_prices.index[-1]
        last_year_idx = df_prices.index.get_indexer([(last_day - relativedelta(years=1))], method="nearest")[
            0
        ]  # one year from last price date
        last_month_idx = df_prices.index.get_indexer([(last_day - relativedelta(months=1))], method="nearest")[0]
        three_months_idx = df_prices.index.get_indexer([(last_day - relativedelta(months=3))], method="nearest")[0]
        six_months_idx = df_prices.index.get_indexer([(last_day - relativedelta(months=6))], method="nearest")[0]
        ytd_idx = df_prices.index.get_indexer([(last_day - BYearEnd())], method="nearest")[0]
        performances = pd.Series(name="performance", dtype=float)

        performances["1_month"] = df_prices.iloc[-1]["close"] / df_prices.iloc[last_month_idx]["close"] - 1
        performances["3_months"] = df_prices.iloc[-1]["close"] / df_prices.iloc[three_months_idx]["close"] - 1
        performances["6_months"] = df_prices.iloc[-1]["close"] / df_prices.iloc[six_months_idx]["close"] - 1
        performances["12_months"] = df_prices.iloc[-1]["close"] / df_prices.iloc[last_year_idx]["close"] - 1
        performances["ytd"] = df_prices.iloc[-1]["close"] / df_prices.iloc[ytd_idx]["close"] - 1
        if df_prices.index[ytd_idx] > df_prices.index[last_month_idx]:
            performances = performances.reindex(["ytd", "1_month", "3_months", "6_months", "12_months"])
        elif df_prices.index[ytd_idx] > df_prices.index[three_months_idx]:
            performances = performances.reindex(["1_month", "ytd", "3_months", "6_months", "12_months"])
        elif df_prices.index[ytd_idx] > df_prices.index[six_months_idx]:
            performances = performances.reindex(["1_month", "3_months", "ytd", "6_months", "12_months"])
        else:
            performances = performances.reindex(["1_month", "3_months", "6_months", "ytd", "12_months"])
        colors = performances.copy()
        colors.loc[performances >= 0], colors.loc[performances < 0] = "green", "darkred"

        performances = performances.mul(100).round(2)
        fig = go.Figure(
            go.Bar(
                x=self._make_presentable_fields(fields_list=performances.index.tolist(), bold=False),
                y=performances.values,
                texttemplate="%{y:.2f}%",
                marker={"color": colors},
            )
        )
        fig.update_yaxes(ticksuffix="%")
        fig.update_layout(title=f'Last Price Date: {last_day.strftime("%Y-%m-%d")}')
        return fig

    def price_and_volume_chart(
        self,
        from_date: dt.date | None = None,
        to_date: dt.date | None = None,
        benchmarks: Optional[list[str]] = None,
        normalize: bool = False,
        short_sma: Optional[int] = 50,
        long_sma: Optional[int] = 200,
        candle_chart: bool = False,
        overlay_volume: bool = False,
    ) -> go.Figure:
        pd.options.plotting.backend = "plotly"
        df_prices = self.get_instrument_price_data(from_date=from_date, to_date=to_date).sort_index()
        if df_prices.empty:
            return go.Figure()

        df_prices = df_prices.asfreq("B")
        df_prices["short_sma"] = df_prices.close.rolling(short_sma).mean() if short_sma else pd.NA
        df_prices["long_sma"] = df_prices.close.rolling(long_sma).mean() if long_sma else pd.NA
        df_prices["volume_diff"] = df_prices.volume.diff()
        df_prices = df_prices.loc[from_date:to_date]

        bar_chart_name = "Volume"
        df_prices["color_volume"] = "green"
        if self.instrument.is_managed:
            bar_chart_name = "Inflow/Outflow"
            df_prices.loc[df_prices.volume < 0, "color_volume"] = "red"
        else:
            perf = df_prices.close.pct_change()
            df_prices.loc[perf < 0, "color_volume"] = "red"
        if overlay_volume:
            fig = make_subplots(specs=[[{"secondary_y": True}]])
        else:
            fig = make_subplots(rows=2, cols=1, shared_xaxes=True, vertical_spacing=0.03, row_width=[0.2, 0.7])
        if not candle_chart:
            instrument_prices = df_prices.loc[:, "close"].dropna()
            cumulative_returns = instrument_prices.pct_change().add(1).cumprod().sub(1).mul(100).fillna(0)
            subfig1 = go.Scatter(
                x=instrument_prices.round(2).index,
                y=instrument_prices.round(2).values,
                mode="lines",
                name=f"{self.instrument.name_repr} - Close Price",
                customdata=cumulative_returns,
                hovertemplate=(
                    f"<b>{self.instrument.name_repr} - Close Price</b>"
                    + "<br>%{x}<br>%{y}<br><i>Cumulative Return</i>: %{customdata:.2f}%<extra></extra>"
                ),
                fill="tozeroy",
                fillcolor="rgba(0,0,255,0.15)",
            )
        else:
            subfig1 = go.Candlestick(
                x=df_prices.index,
                open=df_prices["open"].dropna(),
                high=df_prices["high"].dropna(),
                low=df_prices["low"].dropna(),
                close=df_prices["close"].dropna(),
                name=f"{self.instrument.name_repr}",
            )

        fig.add_trace(subfig1, row=1, col=1) if not overlay_volume else fig.add_trace(subfig1, secondary_y=False)

        for sma in ["short_sma", "long_sma"]:
            name_str = f"Short SMA ({short_sma} days)" if sma == "short_sma" else f"Long SMA ({long_sma} days)"
            sma_fig = go.Scatter(
                x=df_prices.loc[:, sma].dropna().round(2).index,
                y=df_prices.loc[:, sma].dropna().round(2).values,
                mode="lines",
                name=name_str,
                hovertemplate=f"<b>{name_str}</b>" + "<br>%{x}<br>%{y}<extra></extra>",
            )
            fig.add_trace(sma_fig, row=1, col=1) if not overlay_volume else fig.add_trace(sma_fig, secondary_y=False)

        benchmarks = [] if not benchmarks else benchmarks
        qs_benchmarks = (
            InstrumentPrice.objects.filter(
                instrument__in=benchmarks,
                date__range=[from_date, to_date],
                calculated=False,
            )
            .order_by("date", "instrument")
            .annotate(
                fx_rate=CurrencyFXRates.get_fx_rates_subquery_for_two_currencies(
                    "date", "instrument__currency", self.instrument.currency
                ),
                convert_value=ExpressionWrapper(F("net_value") * F("fx_rate"), output_field=FloatField()),
            )
            .values("date", "instrument__name_repr", "convert_value")
        )
        if qs_benchmarks.exists():
            df_benchmarks = pd.DataFrame(qs_benchmarks).set_index(["date", "instrument__name_repr"])
            df_benchmarks = df_benchmarks.convert_value.unstack("instrument__name_repr").astype(float).asfreq("B")
            for benchmark in df_benchmarks.columns:
                df_benchmark = df_benchmarks.loc[:, benchmark].dropna()
                benchmark_cumulative_returns = df_benchmark.pct_change().add(1).cumprod().sub(1).mul(100).fillna(0)
                if normalize and df_prices.index[0] and df_prices.index[0] <= df_benchmark.index[0]:
                    mul_factor = df_prices.loc[df_benchmark.index[0], "close"] / df_benchmark.iat[0]
                    df_benchmark *= mul_factor
                benchmark_fig = go.Scatter(
                    x=df_benchmark.index,
                    y=df_benchmark.round(2).values,
                    mode="lines",
                    name=benchmark,
                    customdata=benchmark_cumulative_returns,
                    hovertemplate=(
                        f"<b>{benchmark}</b>"
                        + "<br>%{x}<br>%{y}<br><i>Cumulative Return</i>: %{customdata:.2f}%<extra></extra>"
                    ),
                )
                (
                    fig.add_trace(benchmark_fig, row=1, col=1)
                    if not overlay_volume
                    else fig.add_trace(benchmark_fig, secondary_y=False)
                )
        subfig2 = go.Bar(
            x=df_prices.loc[:, "close"].dropna().round(2).index,
            y=df_prices.loc[:, "volume"].dropna().round(2).values,
            marker={"color": df_prices.color_volume, "opacity": 0.4},
            name=bar_chart_name,
            hovertemplate="<b>" + bar_chart_name + "</b><br>%{x}<br>%{y:.4s}<extra></extra>",
        )
        fig.add_trace(subfig2, row=2, col=1) if not overlay_volume else fig.add_trace(subfig2, secondary_y=True)

        symbol = self.instrument.currency.symbol
        if overlay_volume:
            fig.update_yaxes(ticksuffix=symbol if symbol else "", title="Price", secondary_y=False)
            fig.update_yaxes(tickformat=".3s", title="Volume", showgrid=False, secondary_y=True)
        else:
            fig.update_yaxes(ticksuffix=symbol if symbol else "", title="Price", row=1, col=1)
            fig.update_yaxes(tickformat=".3s", title="Volume", row=2, col=1)
        fig.update_yaxes(spikesnap="cursor")
        fig.update_xaxes(spikesnap="cursor")
        fig.update_layout(xaxis_rangeslider_visible=False)
        fig.update_layout(showlegend=False)
        fig.update_layout(yaxis=dict(range=[df_prices.close.min(), df_prices.close.max()]))
        return fig

    def instrument_vs_benchmark_prices_chart(
        self,
        benchmark: Optional[Instrument] = None,
        from_date: dt.date | None = None,
        to_date: dt.date | None = None,
    ) -> go.Figure:
        df_prices = self.get_instrument_price_data(from_date=from_date, to_date=to_date)
        if df_prices.empty:
            return go.Figure()
        pd.options.plotting.backend = "plotly"
        df_prices = df_prices.asfreq("B").close.to_frame()
        df_prices = df_prices.loc[from_date:to_date]
        if benchmark and (benchmark_prices := benchmark.prices.filter(date__gte=from_date, date__lte=to_date)):
            df_benchmark_prices = pd.DataFrame(benchmark_prices.values("date", "close"))
            df_benchmark_prices = (
                df_benchmark_prices.set_index("date").rename(columns={"close": benchmark.name_repr}).astype(float)
            )
            df_prices = df_prices.join(df_benchmark_prices)
        instrument_name = self.instrument.name_repr
        df_prices = df_prices.rename(columns={"close": instrument_name})
        df_prices = df_prices.pct_change().add(1).cumprod().sub(1).fillna(0)
        fig = df_prices[f"{instrument_name}"].mul(100).plot.line()
        fig.update_traces(
            hovertemplate=f"<b>{instrument_name}</b>" + "<br>%{y:.2f}%<extra></extra>",
        )
        if benchmark and df_prices.columns.isin([f"{benchmark.name_repr}"]).any():
            fig.add_trace(
                go.Scatter(
                    x=df_prices.loc[:, benchmark.name_repr].dropna().index,
                    y=df_prices.loc[:, benchmark.name_repr].mul(100).dropna().values,
                    mode="lines",
                    name=benchmark.name_repr,
                    hovertemplate=f"<b>{benchmark.name_repr}</b>" + "<br>%{y:.2f}%<extra></extra>",
                )
            )
        fig.update_yaxes(title=None, ticksuffix="%")
        fig.update_xaxes(title="Time")
        fig.update_layout(legend_title_text=None, hovermode="x unified")
        return fig

    def cash_flow_analysis_table_chart(self):
        df_table = self.combine_fundamentals_and_forecasts_df()
        if not df_table.empty:
            df_table = df_table.loc[
                [
                    "revenue",
                    "revenue_growth",
                    "ebitda",
                    "ebitda_margin",
                    "ebit",
                    "capital_expenditures",
                    "cash_from_operation",
                    "investment_cash",
                    "free_cash_flow",
                    "interest_expense",
                    "company_tax_rate",
                ],
                :,
            ]

            df_table.loc["depreciation_and_amortization", :] = df_table.loc["ebitda", :] - df_table.loc["ebit", :]
            df_table.loc["operating_cash_flow_growth", :] = df_table.loc["cash_from_operation", :].pct_change()
            df_table.loc["investing_cash_flow_growth", :] = df_table.loc["investment_cash", :].pct_change()
            df_table.loc["capex_/_depreciation_and_amortization", :] = (
                df_table.loc["capital_expenditures", :] / df_table.loc["depreciation_and_amortization", :]
            )
            df_table.loc["unlevered_free_cash_flow", :] = df_table.loc["free_cash_flow", :] + df_table.loc[
                "interest_expense", :
            ] * (1 - df_table.loc["company_tax_rate", :])
            df_table.loc["unlevered_free_cash_flow_growth", :] = df_table.loc[
                "unlevered_free_cash_flow", :
            ].pct_change()
            df_table.loc["unlevered_free_cash_flow_margin", :] = (
                df_table.loc["unlevered_free_cash_flow", :] / df_table.loc["revenue", :]
            )
            df_table.drop(
                [
                    "depreciation_and_amortization",
                    "capital_expenditures",
                    "ebit",
                    "interest_expense",
                    "company_tax_rate",
                    "free_cash_flow",
                ],
                axis=0,
                inplace=True,
            )
        fig = self._make_pretty_table(df=df_table)
        return fig

    def cash_flow_analysis_bar_chart(self):
        fig = go.Figure()
        df_table = self.combine_fundamentals_and_forecasts_df()
        if not df_table.empty:
            df_table.loc["unlevered_free_cash_flow", :] = df_table.loc["free_cash_flow", :] + df_table.loc[
                "interest_expense", :
            ] * (1 - df_table.loc["company_tax_rate", :])
            df_table = df_table.loc[["cash_from_operation", "investment_cash", "unlevered_free_cash_flow"], :]
            df_table.index = self._make_presentable_fields(df_table.index.tolist(), bold=False)
            for variable in df_table.index:
                y = df_table.loc[variable, :].dropna().div(1000000).round(2).values
                fig.add_bar(
                    y=y,
                    x=df_table.loc[variable, :].dropna().index,
                    name=variable,
                    hovertemplate="<b>" + variable + "</b><br>%{x}<br>%{y:.2f} Mio.<extra></extra>",
                    text=y,
                    textposition="auto",
                )
        symbol = self.instrument.currency.symbol
        fig.update_yaxes(title=f"{symbol if symbol else ''} in Million")
        return fig

    def net_debt_and_ebitda_chart(self):
        fig = make_subplots(specs=[[{"secondary_y": True}]])
        df_table = self.combine_fundamentals_and_forecasts_df()
        if not df_table.empty:
            df_table = df_table.loc[["net_debt", "net_debt_to_ebitda_ratio"], :]
            net_debt_values = df_table.loc["net_debt", :].dropna().div(1000000).round(2).values
            fig.add_bar(
                x=df_table.loc["net_debt", :].dropna().index,
                y=net_debt_values,
                name="Net debt",
                hovertemplate="<b>Net Debt</b><br>%{x}<br>%{y:.2f} Mio.<extra></extra>",
                text=net_debt_values,
                textposition="auto",
            )
            fig.add_scatter(
                x=df_table.loc["net_debt_to_ebitda_ratio", :].dropna().index,
                y=df_table.loc["net_debt_to_ebitda_ratio", :].dropna().mul(100).values,
                name="Net debt / EBITDA",
                yaxis="y2",
                hovertemplate="<b>Net debt / EBITDA</b><br>%{x}<br>%{y:.2f}%<extra></extra>",
            )
        symbol = self.instrument.currency.symbol
        fig.update_yaxes(title=f"{symbol if symbol else ''} in Million", secondary_y=False)
        fig.update_yaxes(ticksuffix="%", secondary_y=True)
        return fig



from contextlib import suppress
from datetime import date
from itertools import cycle

import plotly.express as px
from django.utils.functional import cached_property
from plotly import graph_objects as go
from plotly.subplots import make_subplots
from wbcore import viewsets
from wbcore.utils.date import get_date_interval_from_request

from wbfdm.enums import Indicator, MarketDataChartType
from wbfdm.filters import MarketDataChartFilterSet
from wbfdm.models.instruments import Instrument

from .configs import MarketDataChartTitleConfig, PerformanceSummaryChartTitleConfig
from .mixins import InstrumentMixin


class PerformanceSummaryChartViewSet(InstrumentMixin, viewsets.TimeSeriesChartViewSet):
    queryset = Instrument.objects.all()
    title_config_class = PerformanceSummaryChartTitleConfig

    def get_queryset(self):
        return super().get_queryset().filter(id=self.instrument.id)

    def get_plotly(self, queryset):
        today = date.today()
        trace_factory = (
            queryset.first().technical_analysis(from_date=today.replace(year=today.year - 4)).trace_factory()
        )

        fig = go.Figure()
        for trace in trace_factory.performance_summary_trace(bar_options={"color": px.colors.qualitative.T10[0]}):
            fig.add_trace(trace)
        fig.update_layout(
            template="plotly_white",
            yaxis_tickformat="%",
        )
        return fig


class MarketDataChartViewSet(InstrumentMixin, viewsets.TimeSeriesChartViewSet):
    queryset = Instrument.objects.all()
    filterset_class = MarketDataChartFilterSet
    title_config_class = MarketDataChartTitleConfig

    CHART_MAPPING = {
        MarketDataChartType.CLOSE: "close_trace",
        MarketDataChartType.RETURN: "return_trace",
        MarketDataChartType.LOG_RETURN: "log_return_trace",
        MarketDataChartType.DRAWDOWN: "drawdown_trace",
        MarketDataChartType.CANDLESTICK: "candlestick_trace",
        MarketDataChartType.OHLC: "ohlc_trace",
    }
    BENCHMARK_CHART_MAPPING = {
        MarketDataChartType.CLOSE: "close_trace",
        MarketDataChartType.RETURN: "return_trace",
        MarketDataChartType.LOG_RETURN: "log_return_trace",
        MarketDataChartType.DRAWDOWN: "drawdown_trace",
        MarketDataChartType.CANDLESTICK: "close_trace",
        MarketDataChartType.OHLC: "close_trace",
    }
    STATISTIC_MAPPING = {
        MarketDataChartType.CLOSE: "close",
        MarketDataChartType.RETURN: "cum-ret",
        MarketDataChartType.LOG_RETURN: "cum-log-ret",
        MarketDataChartType.DRAWDOWN: "drawdown",
        MarketDataChartType.CANDLESTICK: "close",
        MarketDataChartType.OHLC: "close",
    }
    INDICATOR_MAPPING = {
        Indicator.SMA_50: 50,
        Indicator.SMA_100: 100,
        Indicator.SMA_120: 120,
        Indicator.SMA_200: 200,
    }

    def get_queryset(self):
        return super().get_queryset().filter(id=self.instrument.id)

    def get_benchmark_queryset(self):
        return Instrument.objects.filter(id__in=self.benchmarks_ids)

    @cached_property
    def benchmarks_ids(self) -> list[int]:
        if benchmarks_str := self.request.GET.get("benchmarks", None):
            return benchmarks_str.split(",")
        return []

    def get_parameters(self) -> list:
        return self.request.GET.get("parameters", "").split(";")

    def get_plotly(self, queryset):
        # Parametrization
        chart_type = MarketDataChartType(self.request.GET.get("chart_type", "close"))
        volume = self.request.GET.get("volume", "false") == "true"
        show_estimates = self.request.GET.get("show_estimates", "false") == "true"
        from_date, to_date = get_date_interval_from_request(self.request, date_range_fieldname="period")
        colors = cycle(px.colors.qualitative.T10)

        # Bootstrap chart
        fig = make_subplots(
            rows=2 if volume else 1,
            cols=1,
            shared_xaxes=True,
            vertical_spacing=0.03,
            row_width=[0.2, 0.7] if volume else [1],
        )

        # Generate Chart for main timeseries
        ta = queryset.first().technical_analysis(from_date, to_date)
        factory = ta.trace_factory()
        for trace in getattr(factory, self.CHART_MAPPING[chart_type])(
            line_options=dict(color=next(colors)), show_estimates=show_estimates
        ):
            fig.add_trace(trace)

        # Generate Charts for all added benchmarks
        for benchmark in self.get_benchmark_queryset():
            _factory = benchmark.technical_benchmark_analysis(ta.df.index.min(), to_date).trace_factory()
            for trace in getattr(_factory, self.BENCHMARK_CHART_MAPPING[chart_type])(
                base_series=ta.df[self.STATISTIC_MAPPING[chart_type]],
                line_options=dict(color=next(colors)),
            ):
                fig.add_trace(trace)

        # Generate Charts for any selected indicators
        if indicators := self.request.GET.get("indicators", None):
            for indicator in indicators.split(","):
                for trace in factory.sma_trace(
                    self.INDICATOR_MAPPING[Indicator(indicator)], line_options=dict(color=next(colors))
                ):
                    fig.add_trace(trace)

        if volume:
            fig.add_trace(
                factory.volume_trace(),
                row=2,
                col=1,
            )

        fig.update_layout(
            template="plotly_white",
            legend=dict(x=0.02, y=0.98, bgcolor="rgba(0,0,0,0)"),
            margin=dict(l=0, r=0, t=0, b=40),
            xaxis_rangeslider_visible=False,
            showlegend=True,
            hovermode="x",
        )
        fig.update_xaxes(rangebreaks=[{"pattern": "day of week", "bounds": [6, 1]}])

        for d in fig.data:
            with suppress(AttributeError, IndexError, ValueError):  # Either Candlestick or OHCL
                if (y := d.y[-1]) is not None:
                    text_value = (
                        f"{y:.2%}"
                        if chart_type in [MarketDataChartType.RETURN, MarketDataChartType.LOG_RETURN]
                        else f"{y:.1f}"
                    )
                    fig.add_scatter(
                        x=[d.x[-1]],
                        y=[y],
                        name=d.name,
                        mode="markers+text",
                        text=text_value,
                        textfont=dict(color=d.line.color),
                        textposition="middle right",
                        marker=dict(color=d.line.color, size=12, symbol="circle"),
                        legendgroup=d.name,
                        showlegend=False,
                    )
        return fig



from wbcore.metadata.configs.titles import TitleViewConfig


class InstrumentTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Instruments"

    def get_instance_title(self):
        try:
            obj = self.view.get_object()
            instrument_type = obj.instrument_type.name if obj.instrument_type else "Instrument"
            return f"{instrument_type}: {str(obj)}"
        except AssertionError:
            return "Instrument: {{name}}"

    def get_create_title(self):
        return "New Instrument"


class SubviewInstrumentTitleViewConfig(TitleViewConfig):
    def get_list_title(self):
        return f"{str(self.view.instrument)} - {self.view.SUBVIEW_NAME} "

    def get_delete_title(self) -> str:
        return None

    def get_instance_title(self) -> str:
        return None

    def get_create_title(self) -> str:
        return None



# Generated by Django 5.0.14 on 2025-08-27 07:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbfdm', '0031_exchange_apply_round_lot_size_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='instrumentprice',
            name='outstanding_shares',
            field=models.DecimalField(blank=True, decimal_places=4, help_text='The amount of outstanding share for this instrument', max_digits=16, null=True, verbose_name='Outstanding Shares'),
        ),
    ]



# Generated by Django 5.0.6 on 2024-05-27 08:56

import django.core.serializers.json
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("wbfdm", "0021_delete_instrumentdailystatistics"),
    ]

    operations = [
        migrations.CreateModel(
            name="InstrumentMetric",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("basket_id", models.PositiveIntegerField()),
                ("basket_repr", models.CharField(max_length=256)),
                (
                    "date",
                    models.DateField(
                        blank=True,
                        help_text="If date is null, the metric is considered static",
                        null=True,
                        verbose_name="Metric Date",
                    ),
                ),
                ("key", models.CharField(max_length=255, verbose_name="Metric Key")),
                (
                    "metrics",
                    models.JSONField(
                        default=dict, encoder=django.core.serializers.json.DjangoJSONEncoder, verbose_name="Metrics"
                    ),
                ),
                (
                    "basket_content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="instrument_metrics",
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "instrument",
                    models.ForeignKey(
                        blank=True,
                        help_text="Instrument where this metric belongs to. If null, the metric is applicable to all instruments related to the basket",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metrics",
                        to="wbfdm.instrument",
                    ),
                ),
                (
                    "parent_metric",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="dependent_metrics",
                        to="metric.instrumentmetric",
                        verbose_name="Parent Metric",
                    ),
                ),
            ],
            options={
                "verbose_name": "Instrument Metric",
                "verbose_name_plural": "Instrument Metrics",
                "indexes": [
                    models.Index(
                        fields=["basket_content_type", "basket_id", "key", "date"],
                        name="metric_inst_basket__30dcda_idx",
                    )
                ],
            },
        ),
        migrations.AddConstraint(
            model_name="instrumentmetric",
            constraint=models.UniqueConstraint(
                fields=("basket_content_type", "basket_id", "instrument", "date", "key"),
                name="unique_instrument_metric",
            ),
        ),
    ]



# Generated by Django 5.0.6 on 2024-06-21 12:42

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("io", "0007_alter_exportsource_query_params"),
        ("wbfdm", "0019_alter_controversy_source"),
    ]

    operations = [
        migrations.CreateModel(
            name="OptionAggregate",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("type", models.CharField(choices=[("PUT", "Put"), ("CALL", "Call")], max_length=6)),
                ("date", models.DateField()),
                ("volume", models.FloatField(blank=True, help_text="Option Volume", null=True, verbose_name="Volume")),
                (
                    "volume_5d",
                    models.FloatField(blank=True, help_text="Option Volume (5D)", null=True, verbose_name="Volume 5D"),
                ),
                (
                    "volume_20d",
                    models.FloatField(
                        blank=True, help_text="Option Volume (20D)", null=True, verbose_name="Volume 20D"
                    ),
                ),
                (
                    "volume_50d",
                    models.FloatField(
                        blank=True, help_text="Option Volume (50D)", null=True, verbose_name="Volume 50D"
                    ),
                ),
                (
                    "open_interest",
                    models.FloatField(blank=True, help_text="Open Interest", null=True, verbose_name="Open Interest"),
                ),
                ("volatility_30d", models.FloatField(blank=True, null=True, verbose_name="Volatility (30D)")),
                ("volatility_60d", models.FloatField(blank=True, null=True, verbose_name="Volatility (60D)")),
                ("volatility_90d", models.FloatField(blank=True, null=True, verbose_name="Volatility (90D)")),
                (
                    "import_source",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="io.importsource"
                    ),
                ),
                (
                    "instrument",
                    models.ForeignKey(
                        limit_choices_to=models.Q(("children__isnull", True)),
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="%(class)s",
                        to="wbfdm.instrument",
                        verbose_name="Instrument",
                    ),
                ),
            ],
            options={
                "verbose_name": "Option Aggregate",
                "verbose_name_plural": "Options Aggregates",
            },
        ),
        migrations.CreateModel(
            name="Option",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("type", models.CharField(choices=[("PUT", "Put"), ("CALL", "Call")], max_length=6)),
                ("date", models.DateField()),
                ("volume", models.FloatField(blank=True, help_text="Option Volume", null=True, verbose_name="Volume")),
                (
                    "volume_5d",
                    models.FloatField(blank=True, help_text="Option Volume (5D)", null=True, verbose_name="Volume 5D"),
                ),
                (
                    "volume_20d",
                    models.FloatField(
                        blank=True, help_text="Option Volume (20D)", null=True, verbose_name="Volume 20D"
                    ),
                ),
                (
                    "volume_50d",
                    models.FloatField(
                        blank=True, help_text="Option Volume (50D)", null=True, verbose_name="Volume 50D"
                    ),
                ),
                (
                    "open_interest",
                    models.FloatField(blank=True, help_text="Open Interest", null=True, verbose_name="Open Interest"),
                ),
                ("volatility_30d", models.FloatField(blank=True, null=True, verbose_name="Volatility (30D)")),
                ("volatility_60d", models.FloatField(blank=True, null=True, verbose_name="Volatility (60D)")),
                ("volatility_90d", models.FloatField(blank=True, null=True, verbose_name="Volatility (90D)")),
                ("contract_identifier", models.CharField(max_length=255, verbose_name="Contract Name")),
                ("strike", models.FloatField(verbose_name="Strike")),
                ("expiration_date", models.DateField(verbose_name="Expiration Date")),
                ("open", models.FloatField(blank=True, null=True, verbose_name="Open")),
                ("high", models.FloatField(blank=True, null=True, verbose_name="High")),
                ("low", models.FloatField(blank=True, null=True, verbose_name="Low")),
                ("close", models.FloatField(blank=True, null=True, verbose_name="Close")),
                ("bid", models.FloatField(blank=True, null=True, verbose_name="Bid")),
                ("ask", models.FloatField(blank=True, null=True, verbose_name="Ask")),
                (
                    "vwap",
                    models.FloatField(blank=True, help_text="Open Interest", null=True, verbose_name="Open Interest"),
                ),
                (
                    "risk_delta",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Delta"', null=True, verbose_name="Delta"
                    ),
                ),
                (
                    "risk_theta",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Theta"', null=True, verbose_name="Theta"
                    ),
                ),
                (
                    "risk_gamma",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Gamma"', null=True, verbose_name="Gamma"
                    ),
                ),
                (
                    "risk_vega",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Vega"', null=True, verbose_name="Vega"
                    ),
                ),
                (
                    "risk_rho",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Rho"', null=True, verbose_name="Rho"
                    ),
                ),
                (
                    "risk_lambda",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Lambda"', null=True, verbose_name="Lambda"
                    ),
                ),
                (
                    "risk_epsilon",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Epsilon"', null=True, verbose_name="Epsilon"
                    ),
                ),
                (
                    "risk_vomma",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Vomma"', null=True, verbose_name="Vomma"
                    ),
                ),
                (
                    "risk_vera",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Vera"', null=True, verbose_name="Vera"
                    ),
                ),
                (
                    "risk_speed",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Speed"', null=True, verbose_name="Speed"
                    ),
                ),
                (
                    "risk_zomma",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Zomma"', null=True, verbose_name="Zomma"
                    ),
                ),
                (
                    "risk_color",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Color"', null=True, verbose_name="Color"
                    ),
                ),
                (
                    "risk_ultima",
                    models.FloatField(
                        blank=True, help_text='Option risk metrics "Ultima"', null=True, verbose_name="Ultima"
                    ),
                ),
                (
                    "import_source",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="io.importsource"
                    ),
                ),
                (
                    "instrument",
                    models.ForeignKey(
                        limit_choices_to=models.Q(("children__isnull", True)),
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="%(class)s",
                        to="wbfdm.instrument",
                        verbose_name="Instrument",
                    ),
                ),
            ],
            options={
                "verbose_name": "Option",
                "verbose_name_plural": "Options",
                "indexes": [
                    models.Index(fields=["instrument", "date", "type"], name="wbfdm_optio_instrum_aeea29_idx"),
                    models.Index(fields=["instrument", "date"], name="wbfdm_optio_instrum_1529cf_idx"),
                    models.Index(fields=["type"], name="wbfdm_optio_type_948af8_idx"),
                ],
            },
        ),
        migrations.AddConstraint(
            model_name="option",
            constraint=models.CheckConstraint(
                check=models.Q(("date__week_day__in", [1, 7]), _negated=True), name="wbfdm_option_weekday_constraint"
            ),
        ),
        migrations.AddConstraint(
            model_name="option",
            constraint=models.UniqueConstraint(
                fields=("instrument", "contract_identifier", "date", "type"), name="unique_option"
            ),
        ),
        migrations.AddIndex(
            model_name="optionaggregate",
            index=models.Index(fields=["instrument", "date", "type"], name="wbfdm_optio_instrum_df2778_idx"),
        ),
        migrations.AddIndex(
            model_name="optionaggregate",
            index=models.Index(fields=["instrument", "date"], name="wbfdm_optio_instrum_6ca630_idx"),
        ),
        migrations.AddIndex(
            model_name="optionaggregate",
            index=models.Index(fields=["type"], name="wbfdm_optio_type_4d93d5_idx"),
        ),
        migrations.AddConstraint(
            model_name="optionaggregate",
            constraint=models.CheckConstraint(
                check=models.Q(("date__week_day__in", [1, 7]), _negated=True),
                name="wbfdm_optionaggregate_weekday_constraint",
            ),
        ),
        migrations.AddConstraint(
            model_name="optionaggregate",
            constraint=models.UniqueConstraint(fields=("instrument", "date", "type"), name="unique_option_aggregate"),
        ),
    ]



from wbcore import filters as wb_filters

from wbfdm.models import Exchange


class ExchangeFilterSet(wb_filters.FilterSet):
    class Meta:
        model = Exchange
        fields = {
            "name": ["exact", "icontains"],
            "mic_code": ["exact", "icontains"],
            "operating_mic_code": ["exact", "icontains"],
            "bbg_composite_primary": ["exact"],
            "bbg_composite": ["exact", "icontains"],
            "refinitiv_identifier_code": ["exact", "icontains"],
            "refinitiv_mnemonic": ["exact", "icontains"],
            "country": ["exact"],
            "city": ["exact"],
            "website": ["exact", "icontains"],
            "comments": ["exact", "icontains"],
            "opening_time": ["exact", "gte", "lte"],
            "closing_time": ["exact", "gte", "lte"],
        }



import pytest
from faker import Faker

from ..backends.performances import InstrumentMetricBaseBackend
from ..dto import MetricField, MetricKey

fake = Faker()


@pytest.mark.django_db
class TestMetricKey:
    @pytest.fixture
    def metric_key(self):
        label = fake.word()

        return MetricKey(
            key=label.lower(),
            label=label,
            additional_prefixes=[fake.word()],
            subfields=[
                MetricField(
                    key=label.lower(),
                    label=label,
                    decorators=[{"position": "left", "value": "&", "type": "text"}],
                    serializer_kwargs={"percent": fake.boolean(), "precision": fake.pyint(min_value=1, max_value=4)},
                )
                for label in map(lambda x: fake.word(), range(3))
            ],
        )

    def test_get_fields(self, metric_key):
        fields = list(metric_key.get_fields())
        assert fields[0] == (
            f"{metric_key.key}_{metric_key.subfields[0].key}",
            f"{metric_key.label} {metric_key.subfields[0].label}",
        )
        assert fields[1] == (
            f"{metric_key.key}_{metric_key.subfields[1].key}",
            f"{metric_key.label} {metric_key.subfields[1].label}",
        )
        assert fields[2] == (
            f"{metric_key.key}_{metric_key.subfields[2].key}",
            f"{metric_key.label} {metric_key.subfields[2].label}",
        )

    def test_get_fields_with_prefixed_key(self, metric_key):
        fields = list(metric_key.get_fields(with_prefixed_key=True))
        prefix = metric_key.additional_prefixes[0]
        assert fields[1] == (
            f"{metric_key.key}_{prefix}_{metric_key.subfields[0].key}",
            f"{metric_key.label} {metric_key.subfields[0].label} ({prefix.title()})",
        )
        assert fields[3] == (
            f"{metric_key.key}_{prefix}_{metric_key.subfields[1].key}",
            f"{metric_key.label} {metric_key.subfields[1].label} ({prefix.title()})",
        )
        assert fields[5] == (
            f"{metric_key.key}_{prefix}_{metric_key.subfields[2].key}",
            f"{metric_key.label} {metric_key.subfields[2].label} ({prefix.title()})",
        )

    def test_fields_map(self, metric_key):
        keys = set(map(lambda x: x[0], metric_key.get_fields(with_prefixed_key=True)))
        assert keys == set(metric_key.subfields_filter_map.keys())

    def test_get_serializer_field_attr(self, weekday, metric_key):
        metric_field = metric_key.subfields[0]
        backend = InstrumentMetricBaseBackend(weekday)
        assert backend.get_serializer_field_attr(metric_field)["percent"] == metric_field.serializer_kwargs["percent"]
        assert (
            backend.get_serializer_field_attr(metric_field)["precision"] == metric_field.serializer_kwargs["precision"]
        )
        assert backend.get_serializer_field_attr(metric_field)["decorators"] == metric_field.decorators



from rest_framework.response import Response
from wbcore import viewsets

from wbfdm.models.instruments import Instrument
from wbfdm.serializers import OfficerSerializer
from wbfdm.viewsets.configs.display.officers import OfficerDisplayViewConfig
from wbfdm.viewsets.configs.titles.instruments import SubviewInstrumentTitleViewConfig

from .mixins import InstrumentMixin


class OfficerViewSet(InstrumentMixin, viewsets.ViewSet):
    IDENTIFIER = "wbfdm:instrument-officers"
    SUBVIEW_NAME = "Officers"
    display_config_class = OfficerDisplayViewConfig
    title_config_class = SubviewInstrumentTitleViewConfig
    serializer_class = OfficerSerializer
    permission_classes = []

    def list(self, request, instrument_id):
        queryset = self.get_queryset()
        serializer = self.get_serializer_class()
        serializer = serializer(queryset, many=True)
        return Response({"results": serializer.data})

    def get_queryset(self):
        queryset = Instrument.objects.filter(id=self.instrument.id)
        return queryset.dl.officers()



from datetime import datetime
from io import BytesIO
from typing import Optional

from django.db import models
from pandas.tseries.offsets import BDay
from wbcore.contrib.io.backends import AbstractDataBackend, register

from .mixin import DataBackendMixin
from .utils import Controller

DEFAULT_MAPPING = {
    "SAL1MN": "revenue_y1",
    "SAL2MN": "revenue_y2",
    "SAL3MN": "revenue_y3",
    "SAL4MN": "revenue_y4",
    "SAL5MN": "revenue_y5",
    "GRM1MN": "gross_profit_margin_without_depreciation_y1",
    "GRM2MN": "gross_profit_margin_without_depreciation_y2",
    "GRM3MN": "gross_profit_margin_without_depreciation_y3",
    "GRM4MN": "gross_profit_margin_without_depreciation_y4",
    "GRM5MN": "gross_profit_margin_without_depreciation_y5",
    "NER1MN": "reported_net_profit_y1",
    "NER2MN": "reported_net_profit_y2",
    "NER3MN": "reported_net_profit_y3",
    "NER4MN": "reported_net_profit_y4",
    "NER5MN": "reported_net_profit_y5",
    "INC1MN": "adjusted_net_profit_y1",
    "INC2MN": "adjusted_net_profit_y2",
    "INC3MN": "adjusted_net_profit_y3",
    "INC4MN": "adjusted_net_profit_y4",
    "INC5MN": "adjusted_net_profit_y5",
    "EBD1MN": "ebitda_y1",
    "EBD2MN": "ebitda_y2",
    "EBD3MN": "ebitda_y3",
    "EBD4MN": "ebitda_y4",
    "EBD5MN": "ebitda_y5",
    "EBT1MN": "ebit_y1",
    "EBT2MN": "ebit_y2",
    "EBT3MN": "ebit_y3",
    "EBT4MN": "ebit_y4",
    "EBT5MN": "ebit_y5",
    "NDT1MN": "net_debt_y1",
    "NDT2MN": "net_debt_y2",
    "NDT3MN": "net_debt_y3",
    "NDT4MN": "net_debt_y4",
    "NDT5MN": "net_debt_y5",
    "EVT1MN": "entreprise_value_y1",
    "EVT2MN": "entreprise_value_y2",
    "EVT3MN": "entreprise_value_y3",
    "EVT4MN": "entreprise_value_y4",
    "EVT5MN": "entreprise_value_y5",
    "FCF1MN": "free_cash_flow_y1",
    "FCF2MN": "free_cash_flow_y2",
    "FCF3MN": "free_cash_flow_y3",
    "FCF4MN": "free_cash_flow_y4",
    "FCF5MN": "free_cash_flow_y5",
    "EPS1MN": "eps_y1",
    "EPS2MN": "eps_y2",
    "EPS3MN": "eps_y3",
    "EPS4MN": "eps_y4",
    "EPS5MN": "eps_y5",
    "CAP1MN": "capital_expenditures_y1",
    "CAP2MN": "capital_expenditures_y2",
    "CAP3MN": "capital_expenditures_y3",
    "CAP4MN": "capital_expenditures_y4",
    "CAP5MN": "capital_expenditures_y5",
    "BPS1MN": "expected_book_value_per_share_y1",
    "BPS2MN": "expected_book_value_per_share_y2",
    "BPS3MN": "expected_book_value_per_share_y3",
    "BPS4MN": "expected_book_value_per_share_y4",
    "BPS5MN": "expected_book_value_per_share_y5",
}

PER_SHARE_FIELDS = [
    "EPS1MN",
    "EPS2MN",
    "EPS3MN",
    "EPS4MN",
    "EPS5MN",
    "BPS1MN",
    "BPS2MN",
    "BPS3MN",
    "BPS4MN",
    "BPS5MN",
    "FCF1MN",
    "FCF2MN",
    "FCF3MN",
    "FCF4MN",
    "FCF5MN",
    "GRM1MN",
    "GRM2MN",
    "GRM3MN",
    "GRM4MN",
    "GRM5MN",
]


CURRENCY_BASED_FIELDS = [
    "SAL1MN",
    "SAL2MN",
    "SAL3MN",
    "SAL4MN",
    "SAL5MN",
    "NER1MN",
    "NER2MN",
    "NER3MN",
    "NER4MN",
    "NER5MN",
    "INC1MN",
    "INC2MN",
    "INC3MN",
    "INC4MN",
    "INC5MN",
    "EBD1MN",
    "EBD2MN",
    "EBD3MN",
    "EBD4MN",
    "EBD5MN",
    "EBT1MN",
    "EBT2MN",
    "EBT3MN",
    "EBT4MN",
    "EBT5MN",
    "NDT1MN",
    "NDT2MN",
    "NDT3MN",
    "NDT4MN",
    "NDT5MN",
    "EVT1MN",
    "EVT2MN",
    "EVT3MN",
    "EVT4MN",
    "EVT5MN",
    "CAP1MN",
    "CAP2MN",
    "CAP3MN",
    "CAP4MN",
    "CAP5MN",
    "FCF1MN",
    "FCF2MN",
    "FCF3MN",
    "FCF4MN",
    "FCF5MN",
]


@register("Forecast", provider_key="refinitiv", save_data_in_import_source=False, passive_only=False)
class DataBackend(DataBackendMixin, AbstractDataBackend):
    def __init__(self, import_credential: Optional[models.Model] = None, **kwargs):
        self.controller = Controller(import_credential.username, import_credential.password)

    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: list[str] = None,
        **kwargs,
    ) -> BytesIO:
        execution_date = execution_time.date()
        start = kwargs.get("start", (execution_date - BDay(1)).date())

        fields = list(DEFAULT_MAPPING.keys())
        if obj_external_ids:
            df = self.controller.get_data(
                obj_external_ids,
                fields,
                start,
                execution_date,
                ibes_non_per_share_fields=list(filter(lambda x: x not in PER_SHARE_FIELDS, fields)),
                ibes_currency_based_fields=CURRENCY_BASED_FIELDS,
            )
            if not df.empty:
                content_file = BytesIO()
                df.to_json(content_file, orient="records")
                file_name = (
                    f"forecast_{start:%Y-%m-%d}-{execution_date:%Y-%m-%d}_{datetime.timestamp(execution_time)}.json"
                )
                yield file_name, content_file



from datetime import date
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pandas.core.indexes.base import Index


def rename_period_index_level_to_repr(df):
    """
    given a dataframe with multi index (year, interim, **other), with interim as a index number from 0 - 4, rename the level 1 index to its string representation (e.g. Q for quarter, S for semester, T for trimister and Y for yearly)

    Returns:
        The dataframe whose index level 1 has been renamed according to its period representation

    """

    if "period_type" in df.index.names:
        df = df.reset_index()
        df["interim"] = df["period_type"] + df["interim"].astype(str)
        df.loc[df["interim"] == "Y0", "interim"] = "Y"
        df = df.drop(columns=["period_type"])
        df = df.set_index(["year", "interim"])
    else:
        renamed_index = []
        for row in df.index:
            if row[1] == 0:
                period_repr = "Y"
            else:
                interim_count = (
                    df.drop(0, level=1, errors="ignore")
                    .loc[(row[0], slice(None), slice(None)), :]
                    .index.get_level_values(1)
                    .max()
                )
                if interim_count == 2:
                    period_repr = f"S{row[1]}"
                elif interim_count == 3:
                    period_repr = f"T{row[1]}"
                elif interim_count == 4:
                    period_repr = f"Q{row[1]}"
                else:
                    period_repr = f"P{row[1]}"

            renamed_index.append((row[0], period_repr, *(row[2:] if len(row) > 2 else ())))
        df = df.set_index([renamed_index])
    return df


def rename_date_columns(columns: "Index", fmt: str) -> dict[str, str]:
    column_mapping = dict()
    for col in filter(lambda col: isinstance(col, date), columns):
        column_mapping[col] = col.strftime(fmt)

    return column_mapping



from wbfdm.import_export.backends.refinitiv.geographic_segment import DEFAULT_MAPPING

from .utils import parse_periodic_fundamental_data


def parse(import_source):
    return parse_periodic_fundamental_data(import_source, DEFAULT_MAPPING)



from .esg import InstrumentESGControversiesViewSet, InstrumentESGPAIViewSet
from .instruments import *
from .market_data import MarketDataChartViewSet, PerformanceSummaryChartViewSet
from .prices import InstrumentPriceViewSet
from .officers import OfficerViewSet

from .financial_analysis import *
from .technical_analysis import *
from .statements import *
from .exchanges import *



# Generated by Django 4.2.6 on 2023-10-30 12:39

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0007_instrument_country_instrument_currency_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="Controversy",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("external_id", models.CharField(max_length=512, unique=True)),
                ("headline", models.TextField(verbose_name="Headline")),
                ("description", models.TextField(verbose_name="Description")),
                ("source", models.CharField(max_length=512, verbose_name="Source")),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("ONGOING", "Ongoing"),
                            ("PARTIALLY_CONCLUDED", "Partially Concluded"),
                            ("CONCLUDED", "Concluded"),
                        ],
                        max_length=64,
                        verbose_name="Status",
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        choices=[("STRUCTURAL", "Structural"), ("NON_STRUCTURAL", "Non Structural")],
                        max_length=64,
                        verbose_name="Type",
                    ),
                ),
                (
                    "severity",
                    models.CharField(
                        choices=[
                            ("MINOR", "Minor"),
                            ("MODERATE", "Moderate"),
                            ("SEVERE", "Severe"),
                            ("Very SEVERE", "Very Severe"),
                        ],
                        max_length=64,
                        verbose_name="Severity",
                    ),
                ),
                (
                    "flag",
                    models.CharField(
                        choices=[
                            ("GREEN", "Green"),
                            ("YELLO", "Yellow"),
                            ("ORANGE", "Orange"),
                            ("RED", "Red"),
                            ("UNKNOWN", "Unknown"),
                        ],
                        max_length=64,
                        verbose_name="Flag",
                    ),
                ),
                ("direct_involvement", models.BooleanField(default=True, verbose_name="Direct Involvement")),
                ("company_response", models.CharField(blank=True, max_length=512, null=True)),
                ("review", models.DateField(verbose_name="Reviewed")),
                ("initiated", models.DateField(verbose_name="initiated")),
                ("instrument", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="wbfdm.instrument")),
            ],
        ),
    ]



from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class InstrumentPriceButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.HyperlinkButton(
                key="import_source",
                label="Import Source",
                icon=WBIcon.SAVE.icon,
            )
        }


class InstrumentPriceInstrumentButtonConfig(InstrumentPriceButtonConfig):
    pass


class FinancialStatisticsInstrumentButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self):
        if instrument_id := self.view.kwargs.get("instrument_id", None):
            return {
                bt.DropDownButton(
                    label="Charts",
                    icon=WBIcon.UNFOLD.icon,
                    buttons=(
                        bt.WidgetButton(
                            endpoint=f'{reverse("wbfdm:market_data-list", args=[instrument_id], request=self.request)}?chart_type=ret',
                            label="Returns Chart",
                            icon=WBIcon.CHART_BARS_HORIZONTAL.icon,
                        ),
                        bt.WidgetButton(
                            endpoint=f'{reverse("wbfdm:market_data-list", args=[instrument_id], request=self.request)}?chart_type=drawdown',
                            label="Drawdowns Chart",
                            icon=WBIcon.CHART_BARS_HORIZONTAL.icon,
                        ),
                    ),
                ),
                bt.WidgetButton(
                    key="bestandworstreturns",
                    label="Best and Worst returns table",
                    icon=WBIcon.CHART_BARS_HORIZONTAL.icon,
                ),
            }

        return set()



from decimal import Decimal
from unittest.mock import patch

import pytest
from faker import Faker

from wbfdm.dataloaders.proxies import InstrumentDataloaderProxy

fake = Faker()


@pytest.mark.django_db
class TestInstrumentFinancialStatisticsMetricBackend:
    @patch.object(InstrumentDataloaderProxy, "financials")
    def test_default_daily_statistics(self, mock_fct, weekday, instrument, instrument_price_factory):
        from wbfdm.contrib.metric.backends.statistics import (  # we import locally to avoid database access pytest error
            InstrumentFinancialStatisticsMetricBackend,
        )

        backend = InstrumentFinancialStatisticsMetricBackend(weekday)

        revenue_y_1 = fake.pyfloat()
        revenue_y0 = fake.pyfloat()
        revenue_y1 = fake.pyfloat()
        market_capitalization = fake.pyfloat()
        price = Decimal(150)
        volume_50d = fake.pyfloat()
        instrument_price_factory.create(
            instrument=instrument,
            calculated=False,
            date=weekday,
            net_value=price,
            volume_50d=volume_50d,
            market_capitalization=market_capitalization,
        )
        instrument.update_last_valuation_date()
        mock_fct.return_value = [
            {"year": weekday.year - 1, "value": revenue_y_1, "instrument_id": instrument.id},
            {"year": weekday.year, "value": revenue_y0, "instrument_id": instrument.id},
            {"year": weekday.year + 1, "value": revenue_y1, "instrument_id": instrument.id},
        ]
        metrics = next(backend.compute_metrics(instrument)).metrics

        assert metrics["revenue_y_1"] == revenue_y_1
        assert metrics["revenue_y0"] == revenue_y0
        assert metrics["revenue_y1"] == revenue_y1
        assert metrics["market_capitalization"] == market_capitalization
        assert metrics["price"] == price
        assert metrics["volume_50d"] == volume_50d



from django.db import connections
from wbfdm.contrib.qa.sync.utils import (
    get_item,
    trigger_partial_update,
    update_instruments,
    update_or_create_item,
)
from wbfdm.models import Instrument
from wbfdm.sync.abstract import Sync


class QACompanySync(Sync[Instrument]):
    SOURCE = "qa-ds2-company"

    def update(self, **kwargs):
        update_instruments("qa/sql/companies.sql", **kwargs)

    def update_or_create_item(self, external_id: int) -> Instrument:
        return update_or_create_item(external_id, self.get_item, self.SOURCE)

    def update_item(self, item: Instrument) -> Instrument:
        return self.update_or_create_item(item.source_id)

    def get_item(self, external_id: int) -> dict:
        return get_item(external_id, "qa/sql/companies.sql")

    def trigger_partial_update(self):
        trigger_partial_update(
            Instrument.objects.filter(source=self.SOURCE),
            "last_update",
            "DsCmpyCode",
            "DS2Company_changes",
            self.update,
            self.update_or_create_item,
        )


class QAInstrumentSync(Sync[Instrument]):
    SOURCE = "qa-ds2-security"

    def update(self, **kwargs):
        update_instruments("qa/sql/instruments.sql", parent_source="qa-ds2-company", **kwargs)

    def update_or_create_item(self, external_id: int) -> Instrument:
        return update_or_create_item(external_id, self.get_item, self.SOURCE, "qa-ds2-company")

    def update_item(self, item: Instrument) -> Instrument:
        return self.update_or_create_item(item.source_id)

    def get_item(self, external_id: int) -> dict:
        return get_item(external_id, "qa/sql/instruments.sql")

    def trigger_partial_update(self):
        trigger_partial_update(
            Instrument.objects.filter(source=self.SOURCE),
            "last_update",
            "DsSecCode",
            "DS2Security_changes",
            self.update,
            self.update_or_create_item,
        )


class QAQuoteSync(Sync[Instrument]):
    SOURCE = "qa-ds2-quote"

    def update(self, **kwargs):
        count = connections["qa"].cursor().execute("SELECT COUNT(*) FROM DS2ExchQtInfo").fetchone()[0]
        for offset in range(0, count, 100_000):
            update_instruments(
                "qa/sql/quotes.sql",
                parent_source="qa-ds2-security",
                context={"offset": offset, "batch": 100_000},
                **kwargs,
            )

    def update_or_create_item(self, external_id: int) -> Instrument:
        return update_or_create_item(external_id, self.get_item, self.SOURCE, "qa-ds2-security")

    def update_item(self, item: Instrument) -> Instrument:
        return self.update_or_create_item(item.source_id)

    def get_item(self, external_id: int) -> dict:
        return get_item(external_id, "qa/sql/quotes.sql")

    def trigger_partial_update(self):
        trigger_partial_update(
            Instrument.objects.filter(source=self.SOURCE),
            "last_update",
            "CONCAT(InfoCode, '-', ExchIntCode)",
            "DS2ExchQtInfo_changes",
            self.update,
            self.update_or_create_item,
        )



import json
from contextlib import suppress

from wbcore.contrib.geography.models import Geography

from wbfdm.models import ClassificationGroup

FIELDS_MAP = {
    "orgName": "name",
    "url": "primary_url",
    "description": "description",
}


def parse(import_source):
    cbinsight_group = ClassificationGroup.objects.get_or_create(name="CBinsights", max_depth=2)[0]
    content = json.load(import_source.file)

    data = []
    for org_data in content:
        d = {
            "provider_id": org_data["orgId"],
            "description": org_data["description"],
            "name": org_data["orgName"],
            "primary_url": org_data["url"],
            "instrument_type": "private_equity",
        }
        if summary := org_data.get("orgSummary", None):
            if additional_urls := summary.get("additionalUrls", None):
                d["additional_urls"] = additional_urls
            if alternative_names := summary.get("aliases", None):
                d["alternative_names"] = alternative_names
            d["founded_year"] = summary["foundedYear"]

            if (country_name := summary.get("country")) and (
                country := Geography.countries.get_by_name(country_name.strip())
            ):
                d["country"] = country.id

                if city_name := summary.get("city", None):
                    # We comment this out because looking up the city at import time drastically decrease speed
                    # if city := Geography.cities.get_by_name(city_name.strip(), parent__parent=country):
                    #     d["headquarter_city"] = city.id
                    if (postal_code := summary.get("postalCode", None)) and (street := summary.get("street", None)):
                        d["headquarter_address"] = f"{street}, {postal_code} {city_name}"

            if sector_id := summary.get("sectorId", None):
                with suppress(Exception):
                    code_aggregated = f"{int(sector_id):03}"
                    if industry_id := summary.get("industryId", None):
                        code_aggregated += f"{int(industry_id):03}"
                        if subindustry_id := summary.get("subindustryId", None):
                            code_aggregated += f"{int(subindustry_id):03}"
                d["classifications"] = [{"code_aggregated": code_aggregated, "group": cbinsight_group.id}]
        data.append(d)
    return {"data": data}



from import_export import fields
from import_export.widgets import ForeignKeyWidget
from wbcore.contrib.io.resources import FilterModelResource

from wbfdm.models import Instrument, InstrumentPrice


class InstrumentPriceExportResource(FilterModelResource):
    """
    Instrument Price Resource class to use to export instrument price from the viewset
    """

    instrument = fields.Field(
        column_name="instrument",
        attribute="instrument",
        widget=ForeignKeyWidget(Instrument, field="isin"),
    )

    class Meta:
        fields = (
            "date",
            "net_value",
            "gross_value",
            "market_capitalization",
            "sharpe_ratio",
            "correlation",
            "beta",
            "outstanding_shares_consolidated",
            "volume",
            "volume_50d",
            "instrument",
        )
        export_order = fields
        model = InstrumentPrice



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

INSTRUMENT_MENUITEM = MenuItem(
    label="Instrument",
    endpoint="wbfdm:instrument-list",
    endpoint_get_parameters={"parent__isnull": True},
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbfdm.view_instrument"]
    ),
)

INVESTABLE_UNIVERSE_MENUITEM = MenuItem(
    label="Investable Universe",
    endpoint="wbfdm:instrument-list",
    endpoint_get_parameters={"is_investable_universe": True},
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbfdm.view_instrument"]
    ),
)



import itertools
from typing import Any, Generic, Iterable, TypedDict, TypeVar, cast

from django.core.cache import cache

T = TypeVar("T", bound=TypedDict)


class Cache(Generic[T]):
    """
    A Cache Class to handle 3-Dimensional data (identifier-key-value)
    """

    def __init__(
        self,
        identifier_key: str = "instrument_id",
        symbol_key: str = "factor_code",
        value_key: str = "value",
        timeout: int | None = None,
    ):
        """
        Constructor of the Cache Class
        Args:
            identifier_key: The lookup identifier field. Default to `instrument_id`
            symbol_key: The symbol lookup field. Default to `factor_code`
            value_key: The value lookup field. Default to `value`
            timeout: The cache timeout configuration value. Default to None (never expired)
        """
        self.identifier_key = identifier_key
        self.symbol_key = symbol_key
        self.value_key = value_key
        self.timeout = timeout
        self.id_symbol_pairs: list[tuple[str, str]] = []
        self.missing_keys = []
        self.missing_ids = set()
        self.missing_symbols = set()

    def _get_cache_key(self, id: str, symbol: str) -> str:
        """
        Generate the key used in the caching layer based on the given ID and symbol.

        Args:
            id (str): The ID to be used in the cache key.
            symbol (str): The symbol to be used in the cache key.

        Returns:
            str: A cache key in the format "id_symbol" where both the ID and symbol
            are in lowercase.
        """
        return f"{str(id).lower()}_{symbol.lower()}"

    def _deserialize_cache(self, id: str, symbol: str, value: Any) -> T:
        """
        Data retreived from cache is deserialized using this method. We expect a key value format which is converted into a dictionary with keys as identifier, symbol and value

        Args:
            id: identifier
            symbol: symbol
            value: value

        Returns:
            a dictionary with keys as identifier, symbol and value
        """
        res = {self.identifier_key: int(id), self.symbol_key: symbol, self.value_key: value}
        return cast(T, res)

    def initialize(self, ids: list[int], symbols: list[str]):
        """
        Initialize the instance with a list of IDs and symbols.

        This method takes a list of identifiers and a list of symbols and creates a list of
        tuples containing all possible pairs of identifiers (converted to strings) and symbols.
        It also initializes sets for symbol and identifiers not present in cache

        Args:
            ids (list[int]): A list of integer IDs.
            symbols (list[str]): A list of symbol strings.
        """
        self.id_symbol_pairs = list(map(lambda x: (str(x[0]), x[1]), itertools.product(ids, symbols)))
        self.missing_ids = set()
        self.missing_symbols = set()
        self.missing_keys = list()

    def fetch_from_cache(self) -> Iterable[T]:
        """
        Fetch values already stored in the cache.

        This method iterates over all identifiers and symbols pair and return the cached value if it exists or mark this pair as missing

        Yields:
            T: Deserialized cached values.
        """
        sentinel = object()
        for id, symbol in self.id_symbol_pairs:
            key = self._get_cache_key(id, symbol)
            cached_value = cache.get(key, sentinel)
            if cached_value is not None and cached_value is not sentinel:
                yield self._deserialize_cache(id, symbol, cached_value)
            elif (
                cached_value is sentinel
            ):  # otherwise, it's literal None and then it means the cache contains "None" for that key
                self.missing_symbols.add(symbol)
                self.missing_ids.add(id)
                self.missing_keys.append(key)

    def write(self, row: T) -> T:
        """
        Write given row into the cache

        Args:
            row: a dictionary typed object

        Returns:
            a dictionary typed object
        """
        if (key := row.get(self.identifier_key)) and (symbol := row.get(self.symbol_key)):
            value = row.get(self.value_key)
            key = self._get_cache_key(str(key), str(symbol))
            cache.set(key, value, timeout=self.timeout)
            if key in self.missing_keys:  # mark row's key are "handled"
                self.missing_keys.remove(key)
        return row

    def close(self):
        """
        Close cache stream by fixing missing keys not handled as None value in the cache (it shows that the value was actually fetched but didn't returned any value)
        """
        for key in self.missing_keys:
            cache.set(key, None, timeout=self.timeout)



from .instruments import InstrumentFactory, CashFactory, EquityFactory, InstrumentTypeFactory
from .instrument_prices import InstrumentPriceFactory
from .controversies import ControversyFactory
from .exchanges import ExchangeFactory
from .classifications import (
    ClassificationGroupFactory,
    ClassificationFactory,
    ParentClassificationFactory,
    InstrumentClassificationThroughModelFactory,
)
from .instrument_list import InstrumentListFactory, InstrumentListThroughModelFactory
from .instruments_relationships import RelatedInstrumentThroughModelFactory, InstrumentFavoriteGroupFactory
from .options import OptionFactory, OptionAggregateFactory



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models.signals import m2m_changed
from django.dispatch import receiver
from rest_framework.reverse import reverse
from wbcore.contrib.tags.models import TagModelMixin
from wbcore.models import WBModel
from wbcore.signals import pre_merge


class InstrumentClassificationRelatedInstrument(models.Model):
    class Type(models.TextChoices):
        PARTNER = "PARTNER", "Partner"
        SUPPLIER = "SUPPLIER", "Supplier"
        PEER = "PEER", "Peer"
        COMPETITOR = "COMPETITOR", "Competitor"
        BIGGEST_THREAT = "BIGGEST_THREAT", "Biggest Threat"
        CUSTOMER = "CUSTOMER", "Customer"

    related_instrument_type = models.CharField(max_length=16, choices=Type.choices, null=True, blank=True)
    classified_instrument = models.ForeignKey(
        to="wbfdm.InstrumentClassificationThroughModel",
        related_name="classification_instrument_relationships",
        on_delete=models.CASCADE,
    )
    related_instrument = models.ForeignKey(
        to="wbfdm.Instrument",
        limit_choices_to=(models.Q(instrument_type__is_classifiable=True) & models.Q(level=0)),
        related_name="instrument_classification_related",
        on_delete=models.CASCADE,
    )

    def __str__(self) -> str:
        return f"{self.classified_instrument} {self.related_instrument}"


class InstrumentClassificationThroughModel(TagModelMixin, models.Model):
    instrument = models.ForeignKey(
        "wbfdm.Instrument",
        on_delete=models.CASCADE,
        related_name="classifications_through",
        limit_choices_to=(models.Q(instrument_type__is_classifiable=True) & models.Q(level=0)),
    )
    classification = models.ForeignKey(
        "wbfdm.Classification",
        on_delete=models.CASCADE,
        related_name="instruments_through",
    )
    is_favorite = models.BooleanField(default=False)
    reason = models.TextField(default="", blank=True, verbose_name="Reason for the choice")
    pure_player = models.BooleanField(default=False, help_text="Pure Players Companies", verbose_name="Pure Player")
    top_player = models.BooleanField(default=False, help_text="Top Players Companies", verbose_name="Top Player")
    percent_of_revenue = models.DecimalField(
        decimal_places=4,
        max_digits=5,
        null=True,
        blank=True,
        verbose_name="% of revenue",
        validators=[MinValueValidator(0), MaxValueValidator(1)],
    )
    related_instruments = models.ManyToManyField(
        to="wbfdm.Instrument",
        limit_choices_to=(models.Q(instrument_type__is_classifiable=True) & models.Q(level=0)),
        through=InstrumentClassificationRelatedInstrument,
        through_fields=("classified_instrument", "related_instrument"),
        blank=True,
    )

    def get_tag_detail_endpoint(self):
        return reverse("wbfdm:classifiedinstrument-detail", [self.id])

    def get_tag_representation(self):
        return f"{self.instrument} - {self.classification}"

    class Meta:
        constraints = [
            models.UniqueConstraint(name="unique_classifiedinstruments", fields=["instrument", "classification"])
        ]

    def __str__(self) -> str:
        return f"{self.instrument} {self.classification}"

    def save(self, *args, **kwargs):
        if self.pure_player and not self.percent_of_revenue:
            self.percent_of_revenue = 1
        return super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbfdm:instrumentclassificationrelationship"


class InstrumentFavoriteGroup(WBModel):
    name = models.CharField(max_length=256)
    instruments = models.ManyToManyField(
        "wbfdm.Instrument",
        related_name="favorite_groups",
        blank=True,
        verbose_name="Favorite Instruments Group",
        limit_choices_to=models.Q(children__isnull=True),
    )
    owner = models.ForeignKey(
        "directory.Person", on_delete=models.CASCADE, blank=True, null=True, related_name="favorite_instruments_groups"
    )
    public = models.BooleanField(default=False, help_text="If set to True, this group will be available to everyone.")
    primary = models.BooleanField(
        default=False,
        help_text="If set to True, this group will be set as default filter for instrument based viewset (Only one primary group allowed).",
    )

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if self.primary:
            InstrumentFavoriteGroup.objects.filter(owner=self.owner, primary=True).exclude(id=self.id).update(
                primary=False
            )
        return super().save(*args, **kwargs)

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbfdm:favoritegroup-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}} - (Public: {{public}})"

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbfdm:favoritegroup"


class RelatedInstrumentThroughModel(models.Model):
    class RelatedTypeChoices(models.TextChoices):
        BENCHMARK = "BENCHMARK", "Benchmark"
        PEER = "PEER", "Peer"
        RISK_INSTRUMENT = "RISK_INSTRUMENT", "Risk Instrument"
        TRACKER = "TRACKER", "Tracker"

    instrument = models.ForeignKey(
        "wbfdm.Instrument",
        on_delete=models.CASCADE,
        related_name="related_instruments_through",
        limit_choices_to=models.Q(children__isnull=True),
    )
    related_instrument = models.ForeignKey(
        "wbfdm.Instrument",
        on_delete=models.CASCADE,
        related_name="dependent_instruments_through",
        limit_choices_to=models.Q(children__isnull=True),
    )
    is_primary = models.BooleanField(default=False)
    related_type = models.CharField(
        max_length=32, default=RelatedTypeChoices.BENCHMARK, choices=RelatedTypeChoices.choices
    )

    class Meta:
        unique_together = ("instrument", "related_instrument", "is_primary", "related_type")

    def __str__(self) -> str:
        return f"{self.instrument} - {self.related_instrument} ({self.related_type})"

    def save(self, *args, **kwargs):
        qs = RelatedInstrumentThroughModel.objects.filter(
            instrument=self.instrument, related_type=self.related_type, is_primary=True
        ).exclude(id=self.id)
        if self.is_primary:
            qs.update(is_primary=False)
        elif not qs.exists():
            self.is_primary = True
        return super().save(*args, **kwargs)


@receiver(m2m_changed, sender="wbfdm.RelatedInstrumentThroughModel")
def add_related_instrument(sender, instance, action, pk_set, **kwargs):
    if action == "post_add" and pk_set:
        for related_instrument_id in pk_set:
            through = RelatedInstrumentThroughModel.objects.filter(
                instrument=instance.id, related_instrument=related_instrument_id
            ).first()
            through.save()
    if action == "post_remove" and pk_set:
        qs = RelatedInstrumentThroughModel.objects.filter(instrument=instance, related_type=instance.related_type)
        if not qs.filter(is_primary=True).exists():
            instance = qs.first()
            instance.is_primary = True
            instance.save()


@receiver(m2m_changed, sender="wbfdm.InstrumentClassificationThroughModel")
def add_classification(sender, instance, action, pk_set, **kwargs):
    if action == "post_add" and pk_set:
        for classification_id in pk_set:
            if through := InstrumentClassificationThroughModel.objects.filter(
                instrument=instance.id, classification=classification_id
            ).first():
                through.save()


@receiver(pre_merge, sender="wbfdm.Instrument")
def pre_merge_instrument(sender: models.Model, merged_object, main_object, **kwargs):
    """
    Reassign all merged instrument preferred classification relationship to the main instrument
    """
    # For every favorite group where the merged instrument is present, we remove it and assign the main instrument instad
    for favorite_group in InstrumentFavoriteGroup.objects.filter(instruments=merged_object):
        favorite_group.instruments.remove(merged_object)
        favorite_group.instruments.add(main_object)

    # For all related instruments relationship of the merged instrument, we reassign them to the main instrument if they don't exist yet. The relationship is then deleted.
    for through in RelatedInstrumentThroughModel.objects.filter(instrument=merged_object):
        RelatedInstrumentThroughModel.objects.get_or_create(
            instrument=main_object,
            related_instrument=through.related_instrument,
            is_primary=through.is_primary,
            related_type=through.related_type,
        )
        through.delete()
    # We also reassign the reverse related instrument relationship where the merged instrument is the related instrument.
    for through in RelatedInstrumentThroughModel.objects.filter(related_instrument=merged_object):
        RelatedInstrumentThroughModel.objects.get_or_create(
            instrument=through.instrument,
            related_instrument=main_object,
            is_primary=through.is_primary,
            related_type=through.related_type,
        )
        through.delete()

    # For all classification relationships of the merged instrument, we reassign them to the main instrument if they don't exist yet. The relationship is then deleted.
    for through in InstrumentClassificationThroughModel.objects.filter(instrument=merged_object):
        new_rel, created = InstrumentClassificationThroughModel.objects.get_or_create(
            instrument=main_object,
            classification=through.classification,
            defaults={
                "is_favorite": through.is_favorite,
                "reason": through.reason,
                "pure_player": through.pure_player,
                "top_player": through.top_player,
                "percent_of_revenue": through.percent_of_revenue,
            },
        )
        for related_instrument in through.related_instruments.all():
            if related_instrument not in new_rel.related_instruments.all():
                new_rel.related_instruments.add(related_instrument)
        through.delete()

    # We also reassign the reverse classification relationship where the merged instrument is the related instrument.
    for through in InstrumentClassificationThroughModel.objects.filter(related_instruments=merged_object):
        through.related_instruments.remove(merged_object)
        through.related_instruments.add(main_object)
